import pandas as pd
import matplotlib.pyplot as plt
import shutil


def load_data(path):
    df = pd.read_json(path)
    return df


def get_test_status(output):
    if 'PASS' in output:
        return 'PASS'
    elif 'FAIL' in output:
        return 'FAIL'
    elif 'SKIP' in output:
        return 'SKIP'
    elif 'ERROR' in output:
        return 'ERROR'
    return 'CRASH'


def get_test_outcome(output):
    output = output.strip()
    if 'PASS' in output:
        return 'Test succeeded'
    elif ('FAIL' in output) or ('SKIP' in output) or ('ERROR' in output):
       return ''.join(output.split('Test results:')[1].strip().split('"')[-1].strip().split('\n')[:-1])
    return 'CRASH'


def get_test_outcome_class(status, outcome, error_message):
    if status == 'FAIL':
        if "Plugin returned incorrect return value" in outcome:
            return f'{status}: Plugin returned incorrect return value x expected y'
    elif status == 'ERROR':
        # ubpf
        if 'at PC' in outcome:
            return f"{status}: {outcome.split('PC')[0]}PC k"
        # windows
        elif "doesn't support helper id=" in outcome:
            return f"{status}: {outcome.split('=')[0]}=k"
        elif 'invalid jump target at offset' in outcome:
            return f"{status}: {outcome.split('offset')[0]}offset k"
        else:
            return f'{status}: {outcome}'
    elif status == 'CRASH':
        error_message = error_message.replace('Unhandled Exception reached the top of main:', '')
        error_message = error_message.replace(', application will now exit', '')
        error_message = error_message.split(':')[0].strip()
        return f'{status}: {error_message}'
    
    # status in ['PASS', 'SKIP']
    return f'{status}: {outcome}'
    

def get_expected_value(output):
    output = output.strip()
    if 'Plugin returned incorrect return value' in output:
        return output.split('expected')[-1].strip()
    return 'NA'


def get_return_value(output):
    output = output.strip()
    if 'Plugin returned incorrect return value' in output:
        return output.split('expected')[0].split('return value')[-1].strip()
    return 'NA'


def preprocess_logs(data):
    print(data['test_filename'])
    # get instruction name
    data['instruction'] = data['test_filename'].apply(lambda x: '_'.join(x.split('_')[:-1]))

    # test status (PASS, FAIL, SKIP)
    data['test_status'] = data['output'].apply(lambda x: get_test_status(x))

    # test outcome 
    data['test_outcome'] = data['output'].apply(lambda x: get_test_outcome(x))
    
    data['test_outcome_class'] = data.apply(lambda x: get_test_outcome_class(x['test_status'], x['test_outcome'], x['error_message']), axis='columns')
    
    data['expected_value'] = data['test_outcome'].apply(lambda x: get_expected_value(x))

    data['return_value'] = data['test_outcome'].apply(lambda x: get_return_value(x))

    return data


def plot_test_status_distribution_instruction(data, output_path, baseline):
    # Define colors for each test_status
    
    colors = {
        'PASS': '#56B4E9',
        'FAIL': '#0072B2',
        'SKIP': '#CC79A7',
        'ERROR': '#D55E00',
        'CRASH': '#E69F00'
    }

    
    data['instruction'] = data['instruction'].apply(lambda x: x.split('_')[0])

    # Group by 'instruction' and count 'test_status'
    counts = data.groupby(['instruction', 'test_status']).size().unstack(fill_value=0)
    counts['sum'] = counts['CRASH'] + counts['ERROR'] + counts['SKIP'] + counts['PASS'] + counts['FAIL'] 
    counts['CRASH'] = 100 * counts['CRASH'] / counts['sum']
    counts['ERROR'] = 100 * counts['ERROR'] / counts['sum']
    counts['SKIP'] = 100 * counts['SKIP'] / counts['sum']
    counts['PASS'] = 100 * counts['PASS'] / counts['sum']
    counts['FAIL'] = 100 * counts['FAIL'] / counts['sum']
    
    counts['invalid'] = counts['CRASH'] + counts['ERROR'] + counts['SKIP']
    counts = counts.sort_values(['invalid'])
    counts.drop(columns=['sum', 'invalid'], inplace=True)
    print(counts)
    counts = counts[['CRASH', 'ERROR', 'SKIP', 'FAIL', 'PASS']]
    print(counts)
    
    # Plotting the stacked bar chart with specified colors
    ax = counts.plot(kind='bar', stacked=True, figsize=(14, 6), color=[colors[col] for col in counts.columns])
    
    # Adding labels and title
    ax.legend(title='Test Status', fontsize=14, title_fontsize=16) 
    ax.set_xlabel('Instruction',fontsize=18)
    ax.set_ylabel('Normalized Test Counts',fontsize=18)
    ax.set_title(f'Test Status vs Instruction (bug-guided-code-diff:windows)', fontsize=20)
    ax.tick_params(axis='both', which='major', labelsize=16)
    plt.tight_layout()
    plt.savefig(f'{output_path}/{baseline}_test_status_vs_instruction.png')


def get_summary_stats(data, output_path, baseline):
    test_status = data[['test_status', 'index']].groupby("test_status").count()
    print('Distribution of test status')
    print(test_status)
    test_status.to_csv(f'{output_path}/{baseline}_test_status.csv')

    test_outcome = data[['test_outcome', 'index']].groupby("test_outcome").count()
    print('Distribution of test outcome')
    print(test_outcome)
    test_outcome.to_csv(f'{output_path}/{baseline}_test_outcome.csv')

    test_outcome_class = data[['test_outcome_class', 'index']].groupby("test_outcome_class").count()
    print('Distribution of test outcome class')
    print(test_outcome_class)
    test_outcome_class.to_csv(f'{output_path}/{baseline}_test_outcome_class.csv')

    error_message = data[['error_message', 'index']].groupby("error_message").count()
    print('Distribution of error_message')
    print(error_message)
    error_message.to_csv(f'{output_path}/{baseline}_error_message.csv')

    

def get_stats_per_instruction(data, output_path, baseline):
    data['instruction'] = data['instruction'].apply(lambda x: x.split('_')[0])
    
    instruction_test_status = data.groupby(['instruction', 'test_status']).size().unstack(fill_value=0)
    print('Distribution of test status per instruction')
    print(instruction_test_status)
    instruction_test_status.to_csv(f'{output_path}/{baseline}_instruction_test_status.csv')
    
    instruction_test_outcome = data.groupby(['instruction', 'test_outcome']).size().unstack(fill_value=0)
    print('Distribution of test outcome per instruction')
    print(instruction_test_outcome)
    instruction_test_outcome.to_csv(f'{output_path}/{baseline}_instruction_test_outcome.csv')

    instruction_error_message = data.groupby(['instruction', 'error_message']).size().unstack(fill_value=0)
    print('Distribution of error_message per instruction')
    print(instruction_error_message)
    instruction_error_message.to_csv(f'{output_path}/{baseline}_instruction_error_message.csv')


def group_testfiles_by_status(data, src_path, dst_path, baseline, target):

    for i, row in data.iterrows():
        print(row)
        src = f'{src_path}/{baseline}/processed/{row.test_filename}'
        dst = f'{dst_path}/{target}/{row.test_status}/{baseline}_{row.test_filename}'
        print(f'scr: {src}')
        print(f'dst: {dst}')
        shutil.copyfile(src, dst)
        print('\n')



def main(baseline, target):

    path = f'./output_logs/output_exec_tests_{baseline}_{target}.json'
    
    output_path = './output'

    print(f'baseline: {baseline}_{target}')
    #data = load_data(path)
    data = pd.read_json(path)
    data.to_json(path, orient='records')
    
    #data = data.dropna()
    data = preprocess_logs(data)
    data.to_json(f'./output_logs/processed_output_{baseline}_{target}.json', orient='records')
    
    print(data)
    
    plot_test_status_distribution_instruction(data, output_path, f'{baseline}_{target}')
    
    get_summary_stats(data, output_path, f'{baseline}_{target}')
    
    get_stats_per_instruction(data, output_path, f'{baseline}_{target}')
  



if __name__ == "__main__":
  
    baselines = ['3shot_random', 'target_section', 'random_instruction_permutation_3', 'random_instruction_permutation_5', 'random_instruction_permutation_10',  'prompt_chain', 'prompt_chain_instruct', 'bug_centric', 'code_description', 'code_description_diff', 'code_diff', 'bug_guided_code_diff']
    
    targets = ['ubpf', 'linux', 'windows']
    
    for target in targets:
        for baseline in baselines:
            main(baseline, target)
    
# to run 
# python analyse_logs.py

