import os
import sys
import time
import random
sys.path.append('../')
from utils.data_utils import load_file, load_as_df, save_file, load_pdf, fetch_relevant_documentation, format_and_save_outputs
from utils.model_utils import prompt_model, is_valid_model_output, get_output_counts_per_instruction


# prompt model 
def prompt_model_single_instruction(query_instructions, manual, examples, model):    
    
    prompt = f"Given a description of the instruction class extracted from the BPF instruction set architecture (ISA), your task is to generate a diverse set of tests that includes all of the following instructions: {query_instructions}. Here is the relevant documentation:\n{manual}. Here are few example tests:\n{examples}. Generate 10 more tests that use a complex combination of the given set of instructions. Make sure you generate the output using the same format and include a description of what the test does. Use <SEP> tag to separate the tests. The longer the test the better."

    # prompt model
    response = prompt_model(model, prompt)
    
    return response


def get_examples_context(test_names, path):
    examples_context = ''
    for test_name in test_names:
        test = load_file(f'{path}/{test_name.strip()}.data')
        examples_context += f'# test: {test_name.strip()}.data\n'
        examples_context += test + '\n\n'
    return examples_context


def get_random_permutation_instructions(instructions_list, k):
    return random.sample(instructions_list, k)


def main():
    start_timer = time.time()

    # load variables
    instructions_df = load_as_df('../context/instructions_ground_truth.txt')
    manual = load_pdf('../context/BPFInstructionSetArchitecture.pdf')
    examples = load_file('../examples/random_3shot.txt')


    num_generations = 10
    num_instructions = 10
    num_permutations = 50

    output_path = f'../../data/all_generated_tests/random_instruction_permutation_{num_instructions}'

    # load model
    model = "gpt-4o"

    print('----------------<output>-------------------')
    
    # random 3-shot examples for all instructions 
    for i in range(num_permutations):
        start = time.time()
        instruction_seq = get_random_permutation_instructions(list(instructions_df['instruction']), num_instructions)

        existing_files = [file for file in os.listdir(f'{output_path}/raw_output/')]

        if any(f'{"_".join(instruction_seq)}_' in file for file in existing_files):
            print(f'{i}: {"_".join(instruction_seq)} already processed | generated tests path: {output_path}/raw_output/{"_".join(instruction_seq)}_10.txt')
            continue

        if len(existing_files) == num_permutations:
            print(f'permutation limit {num_permutations} reached!')
            break
        
        # get relevant sections from documentation
        relevant_documentation = ''
        for inst in instruction_seq:
            reference_section = instructions_df[instructions_df['instruction'] == inst]['reference'].values[0]
            print(inst, reference_section)
            doc = fetch_relevant_documentation(manual, reference_section)
            if doc not in relevant_documentation:
                relevant_documentation += doc + '\n'


        # if model output format is not valid, retry
        while True:
            output = prompt_model_single_instruction(', '.join(instruction_seq), relevant_documentation, examples, model)
            if is_valid_model_output(output, num_generations):
                break
            else:
                print('Invalid model output, prompting llm again')

        # format output
        format_and_save_outputs(output, f'{output_path}/processed', '_'.join(instruction_seq), num_generations)

        end = time.time()

        # log progress 
        print(f"{i}: {'_'.join(instruction_seq)} | generated tests path: {output_path} | time taken: {end - start}")

        

    end_timer = time.time()
    
    print(f'total time: {end_timer - start_timer}')

    get_output_counts_per_instruction(f'{output_path}/processed')



if __name__ == "__main__":
    main()


# to run file
# python -u main_random_instruction_permutation.py 2>&1 | tee output_random_instruction_permutation10.txt
